package edu.washington.cs.atomicdebug.examples;

import edu.washington.cs.atomicdebug.AtomicDebug;

/**
 * The classic Producer-Consumer synchronization problem. A producer thread and
 * two consumer threads share a buffer, and the producer tries to fill the
 * buffer with values, while the consumers try to remove those values and
 * process them. This pattern occurs often in operating systems code.
 * 
 * Our solution uses a busy wait for both producer and consumer, which is
 * suboptimal, but does not have any race conditions, and a true implementation
 * of atomic transactions could eliminate the busy wait using a retry()
 * primitive. There are also many solutions using semaphores and mutexes, or
 * using something like Concurrent ML for functional programming languages.
 * 
 * This "solved" version includes thread sleeps and atomic blocks.
 * 
 * @author Laura Effinger-Dean
 */
public class ProducerConsumerSolved {
	/**
	 * The shared buffer. In general, this could be an array of values.
	 * 
	 * -1 = empty positive number = full
	 */
	private static int buffer = -1;

	/**
	 * The producer tries to fill the buffer with a series of values.
	 */
	private static void produceValues() {
		try {
			for (int i = 0; i < 10; i++) {
				while (true) {
					AtomicDebug.enterAtomic();
					if (buffer == -1) {
						Thread.sleep(10);
						buffer = i * 5 + 3;
						AtomicDebug.exitAtomic();
						break;
					}
					AtomicDebug.exitAtomic();
				}
			}
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private static void consumeValues() {
		try {
			int value;
			for (int i = 0; i < 5; i++) {
				while (true) {
					AtomicDebug.enterAtomic();
					if (buffer != -1) {
						Thread.sleep(10);
						value = buffer;
						Thread.sleep(10);
						buffer = -1;
						AtomicDebug.exitAtomic();
						break;
					}
					AtomicDebug.exitAtomic();
				}
				System.out.println(value);
			}
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static void main(String[] args) {
		Thread producer = new Thread() {
			public void run() {
				produceValues();
			}
		};
		Thread consumer1 = new Thread() {
			public void run() {
				consumeValues();
			}
		};
		Thread consumer2 = new Thread() {
			public void run() {
				consumeValues();
			}
		};
		producer.start();
		consumer1.start();
		consumer2.start();
	}

}
